@@ -5,6 +5,9 @@ import umap.sparse as spdist
 
 from sklearn.metrics import pairwise_distances
 from sklearn.neighbors import BallTree
+from scipy.version import full_version as scipy_full_version
+import pytest
+
 
 
 # ===================================================
@@ -270,6 +273,10 @@ def test_sparse_correlation(sparse_spatial_data):
     sparse_spatial_check("correlation", sparse_spatial_data)
 
 
+def test_sparse_braycurtis(sparse_spatial_data):
+    sparse_spatial_check("braycurtis", sparse_spatial_data)
+
+
 # ---------------------------
 # Sparse Binary Metric Tests
 # ---------------------------
@@ -328,10 +335,12 @@ def test_seuclidean(spatial_data):
         err_msg="Distances don't match " "for metric seuclidean",
     )
 
-
+@pytest.mark.skipif(
+    scipy_full_version < "1.8", reason="incorrect function in scipy<1.8"
+)
 def test_weighted_minkowski(spatial_data):
     v = np.abs(np.random.randn(spatial_data.shape[1]))
-    dist_matrix = pairwise_distances(spatial_data, metric="wminkowski", w=v, p=3)
+    dist_matrix = pairwise_distances(spatial_data, metric="minkowski", w=v, p=3)
     test_matrix = np.array(
         [
             [
@@ -437,7 +446,7 @@ def test_sparse_hellinger(sparse_spatial_data):
         test_matrix,
         dist_matrix,
         err_msg="Sparse distances don't match " "for metric hellinger",
-        decimal=4,
+        decimal=3,
     )
 
     # Ensure ll_dirichlet runs
@@ -484,24 +493,25 @@ def test_grad_metrics_match_metrics(spatial_data, spatial_distances):
         err_msg="Distances don't match " "for metric seuclidean",
     )
 
-    # Weighted minkowski
-    dist_matrix = pairwise_distances(spatial_data, metric="wminkowski", w=v, p=3)
-    test_matrix = np.array(
-        [
+    if scipy_full_version >= "1.8":
+        # Weighted minkowski
+        dist_matrix = pairwise_distances(spatial_data, metric="minkowski", w=v, p=3)
+        test_matrix = np.array(
             [
-                dist.weighted_minkowski_grad(spatial_data[i], spatial_data[j], v, p=3)[
-                    0
+                [
+                    dist.weighted_minkowski_grad(spatial_data[i], spatial_data[j], v, p=3)[
+                        0
+                    ]
+                    for j in range(spatial_data.shape[0])
                 ]
-                for j in range(spatial_data.shape[0])
+                for i in range(spatial_data.shape[0])
             ]
-            for i in range(spatial_data.shape[0])
-        ]
-    )
-    assert_array_almost_equal(
-        test_matrix,
-        dist_matrix,
-        err_msg="Distances don't match " "for metric weighted_minkowski",
-    )
+        )
+        assert_array_almost_equal(
+            test_matrix,
+            dist_matrix,
+            err_msg="Distances don't match " "for metric weighted_minkowski",
+        )
 
     # Mahalanobis
     v = np.abs(np.random.randn(spatial_data.shape[1], spatial_data.shape[1]))
@@ -518,6 +528,7 @@ def test_grad_metrics_match_metrics(spatial_data, spatial_distances):
     assert_array_almost_equal(
         test_matrix,
         dist_matrix,
+        decimal=5,
         err_msg="Distances don't match " "for metric mahalanobis",
     )
 
